#!/usr/bin/env python
""" Convolve ASKAP images to common resolution """
__author__ = "Alec Thomson"

import os
import sys
import numpy as np
import scipy.signal
from astropy import units as u
from astropy.io import fits
from astropy.convolution import convolve, convolve_fft
from radio_beam import Beam, Beams
from radio_beam.utils import BeamError
from . import au2
from . import convolve_uv
import functools
import schwimmbad
import psutil
from tqdm import tqdm
from IPython import embed
import matplotlib.pyplot as plt
import warnings
try:
    print = functools.partial(
        print, f'[{psutil.Process().cpu_num()}]', flush=True)
except AttributeError:
    print = functools.partial(print, flush=True)

#############################################
#### ADAPTED FROM SCRIPT BY T. VERNSTROM ####
#############################################


def round_up(n, decimals=0):
    multiplier = 10 ** decimals
    return np.ceil(n * multiplier) / multiplier


def my_ceil(a, precision=0):
    return np.round(a + 0.5 * 10**(-precision), precision)


def getbeam(datadict, new_beam, cutoff=None, verbose=False):
    """Get beam info
    """
    if verbose:
        print(f"Current beam is", datadict['oldbeam'])

    if cutoff is not None and datadict['oldbeam'].major.to(u.arcsec) > cutoff*u.arcsec:
        return np.nan, np.nan

    else:
        conbm = new_beam.deconvolve(datadict['oldbeam'])
        fac, amp, outbmaj, outbmin, outbpa = au2.gauss_factor(
            [
                conbm.major.to(u.arcsec).value,
                conbm.minor.to(u.arcsec).value,
                conbm.pa.to(u.deg).value
            ],
            beamOrig=[
                datadict['oldbeam'].major.to(u.arcsec).value,
                datadict['oldbeam'].minor.to(u.arcsec).value,
                datadict['oldbeam'].pa.to(u.deg).value
            ],
            dx1=datadict['dx'].to(u.arcsec).value,
            dy1=datadict['dy'].to(u.arcsec).value
        )

        return conbm, fac


def getimdata(cubenm, verbose=False):
    """Get fits image data
    """
    if verbose:
        print(f'Getting image data from {cubenm}')
    with fits.open(cubenm, memmap=True, mode='denywrite') as hdu:

        dxas = hdu[0].header['CDELT1']*-1*u.deg
        dyas = hdu[0].header['CDELT2']*u.deg

        nx, ny = hdu[0].data[0, 0, :,
                             :].shape[0], hdu[0].data[0, 0, :, :].shape[1]

        old_beam = Beam.from_fits_header(
            hdu[0].header
        )

        datadict = {
            'image': hdu[0].data[0, 0, :, :],
            'header': hdu[0].header,
            'oldbeam': old_beam,
            'nx': nx,
            'ny': ny,
            'dx': dxas,
            'dy': dyas
        }
    return datadict


def smooth(datadict, conv_mode='robust', verbose=False):
    """Do the smoothing
    """
    if np.isnan(datadict["sfactor"]):
        if verbose:
            print('Beam larger than cutoff -- blanking')
        newim = np.ones_like(datadict['image']) * np.nan
        return newim
    else:
        # using Beams package
        if verbose:
            print(f'Smoothing so beam is', datadict["final_beam"])
            print(f'Using convolving beam', datadict["conbeam"])
        pix_scale = datadict['dy']

        gauss_kern = datadict["conbeam"].as_kernel(pix_scale)
        conbm1 = gauss_kern.array/gauss_kern.array.max()
        fac = datadict["sfactor"]
        if conv_mode == 'robust':
            newim, fac = convolve_uv.convolve(
                datadict['image'].astype('f8'),
                datadict['oldbeam'],
                datadict['final_beam'],
                datadict['dx'],
                datadict['dy'],
            )
        if conv_mode == 'scipy':
            newim = scipy.signal.convolve(
                datadict['image'].astype('f8'),
                conbm1,
                mode='same'
            )
        elif conv_mode == 'astropy':
            newim = convolve(
                datadict['image'].astype('f8'),
                conbm1,
                normalize_kernel=False,
            )
        elif conv_mode == 'astropy_fft':
            newim = convolve_fft(
                datadict['image'].astype('f8'),
                conbm1,
                normalize_kernel=False,
                allow_huge=True,
            )
        if verbose:
            print(f'Using scaling factor', fac)
        newim *= fac
        return newim


def savefile(datadict, filename, outdir='.', verbose=False):
    """Save file to disk
    """
    outfile = f'{outdir}/{filename}'
    if verbose:
        print(f'Saving to {outfile}')
    header = datadict['header']
    beam = datadict['final_beam']
    header = beam.attach_to_header(header)
    fits.writeto(outfile, datadict['newimage'], header=header, overwrite=True)


def worker(args):
    file, outdir, new_beam, conv_mode, clargs, verbose = args
    if verbose:
        print(f'Working on {file}')

    if outdir is None:
        outdir = os.path.dirname(file)

    if outdir == '':
        outdir = '.'

    outfile = os.path.basename(file)
    outfile = outfile.replace('.fits', f'.{clargs.suffix}.fits')
    if clargs.prefix is not None:
        outfile = clargs.prefix + outfile
    datadict = getimdata(file)

    conbeam, sfactor = getbeam(
        datadict,
        new_beam,
        cutoff=clargs.cutoff,
        verbose=verbose
    )

    datadict.update(
        {
            "conbeam": conbeam,
            "final_beam": new_beam,
            "sfactor": sfactor
        }
    )

    newim = smooth(datadict, conv_mode=conv_mode, verbose=verbose)
    datadict.update(
        {
            "newimage": newim,
        }
    )

    savefile(datadict, outfile, outdir, verbose=verbose)


def getmaxbeam(files, conv_mode='robust', target_beam=None, cutoff=None,
               tolerance=0.0001, nsamps=200, epsilon=0.0005, verbose=False):
    """Get smallest common beam
    """
    beams = []
    for file in files:
        header = fits.getheader(file, memmap=True)
        beam = Beam.from_fits_header(header)
        beams.append(beam)

    beams = Beams(
        [beam.major.value for beam in beams]*u.deg,
        [beam.minor.value for beam in beams]*u.deg,
        [beam.pa.value for beam in beams]*u.deg
    )
    if cutoff is not None:
        flags = beams.major > cutoff*u.arcsec
    else:
        flags = np.array([False for beam in beams])
    try:
        cmn_beam = beams[~flags].common_beam(
            tolerance=tolerance, epsilon=epsilon, nsamps=nsamps)
    except BeamError:
        if verbose:
            print("Couldn't find common beam with defaults")
            print("Trying again with smaller tolerance")
        cmn_beam = beams[~flags].common_beam(
            tolerance=tolerance*0.1, epsilon=epsilon, nsamps=nsamps)

    # Round up values
    cmn_beam = Beam(
        major=my_ceil(cmn_beam.major.to(u.arcsec).value, precision=1)*u.arcsec,
        minor=my_ceil(cmn_beam.minor.to(u.arcsec).value, precision=1)*u.arcsec,
        pa=round_up(cmn_beam.pa.to(u.deg), decimals=2)
    )
    target_header = header
    dx = target_header['CDELT1']*-1*u.deg
    dy = target_header['CDELT2']*u.deg
    if not dx == dy:
        raise Exception("GRID MUST BE SAME IN X AND Y")
    grid = dy
    if conv_mode is not 'robust':
        # Get the minor axis of the convolving beams
        minorcons = []
        for beam in beams:
            minorcons += [cmn_beam.deconvolve(beam).minor.to(u.arcsec).value]
        minorcons = np.array(minorcons)*u.arcsec
        samps = minorcons / grid.to(u.arcsec)
        # Check that convolving beam will be Nyquist sampled
        if any(samps.value < 2):
            # Set the convolving beam to be Nyquist sampled
            nyq_con_beam = Beam(
                major=grid*2,
                minor=grid*2,
                pa=0*u.deg
            )
            # Find new target based on common beam * Nyquist beam
            # Not sure if this is best - but it works
            nyq_beam = cmn_beam.convolve(nyq_con_beam)
            nyq_beam = Beam(
                major=my_ceil(nyq_beam.major.to(
                    u.arcsec).value, precision=1)*u.arcsec,
                minor=my_ceil(nyq_beam.minor.to(
                    u.arcsec).value, precision=1)*u.arcsec,
                pa=round_up(nyq_beam.pa.to(u.deg), decimals=2)
            )
            if verbose:
                print('Smallest common Nyquist sampled beam is:', nyq_beam)
            if target_beam is not None:
                if target_beam < nyq_beam:
                    warnings.warn('TARGET BEAM WILL BE UNDERSAMPLED!')
                    raise Exception("CAN'T UNDERSAMPLE BEAM - EXITING")
            else:
                warnings.warn('COMMON BEAM WILL BE UNDERSAMPLED!')
                warnings.warn('SETTING COMMON BEAM TO NYQUIST BEAM')
                cmn_beam = nyq_beam

    return cmn_beam, beams


def main(pool, args, verbose=False):
    """Main script
    """
    if args.dryrun:
        if verbose:
            print('Doing a dry run -- no files will be saved')
    # Fix up outdir
    outdir = args.outdir
    if outdir is not None:
        if outdir[-1] == '/':
            outdir = outdir[:-1]
    else:
        outdir = None

    # Get file list
    files = sorted(args.infile)
    if files == []:
        raise Exception('No files found!')

    # Parse args

    conv_mode = args.conv_mode
    print(conv_mode)
    if not conv_mode == 'robust' and not conv_mode == 'scipy' and \
            not conv_mode == 'astropy' and not conv_mode == 'astropy_fft':
        raise Exception('Please select valid convolution method!')

    if verbose:
        print(f"Using convolution method {conv_mode}")
        if conv_mode == 'robust':
            print("This is the most robust method. And fast!")
        elif conv_mode == 'scipy':
            print('This fast, but not robust to NaNs or small PSF changes')
        else:
            print('This is slower, but robust to NaNs, but not to small PSF changes')

    bmaj = args.bmaj
    bmin = args.bmin
    bpa = args.bpa

    nonetest = [test is None for test in [bmaj, bmin, bpa]]

    if all(nonetest):
        target_beam = None

    elif not all(nonetest) and any(nonetest):
        raise Exception('Please specify all target beam params!')

    elif not all(nonetest) and not any(nonetest):
        target_beam = Beam(
            bmaj * u.arcsec,
            bmin * u.arcsec,
            bpa * u.deg
        )
        if verbose:
            print('Target beam is ', target_beam)

    # Find smallest common beam
    big_beam, allbeams = getmaxbeam(files,
                                    conv_mode=conv_mode,
                                    target_beam=target_beam,
                                    cutoff=args.cutoff,
                                    tolerance=args.tolerance,
                                    nsamps=args.nsamps,
                                    epsilon=args.epsilon,
                                    verbose=verbose)

    if target_beam is not None:
        if verbose:
            print('Checking that target beam will deconvolve...')

        mask_count = 0
        failed = []
        for i, (beam, file) in enumerate(
            tqdm(
                zip(allbeams, files),
                total=len(allbeams),
                desc='Deconvolving',
                disable=(not verbose)
            )
        ):
            try:
                target_beam.deconvolve(beam)
            except ValueError:
                mask_count += 1
                failed.append(file)
        if mask_count > 0:
            if verbose:
                print('The following images could not reach target resolution:')
                print(failed)
            raise Exception("Please choose a larger target beam!")

        else:
            new_beam = target_beam

    else:
        new_beam = big_beam

    if verbose:
        print(f'Final beam is', new_beam)
    inputs = [[file, outdir, new_beam, conv_mode, args, verbose]
              for i, file in enumerate(files)]

    if not args.dryrun:
        output = list(pool.map(worker, inputs))

    if verbose:
        print('Done!')


def cli():
    """Command-line interface
    """
    import argparse

    # Help string to be shown using the -h option
    descStr = """
    Smooth a field of 2D images to a common resolution.

    Names of output files are 'infile'.sm.fits

    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        'infile',
        metavar='infile',
        type=str,
        help='Input FITS image(s) to smooth (can be a wildcard) - beam info must be in header.',
        nargs='+')

    parser.add_argument(
        '-p',
        '--prefix',
        dest='prefix',
        type=str,
        default=None,
        help='Add prefix to output filenames.')

    parser.add_argument(
        '-s',
        '--suffix',
        dest='suffix',
        type=str,
        default='sm',
        help='Add suffix to output filenames [...sm.fits].')

    parser.add_argument(
        '-o',
        '--outdir',
        dest='outdir',
        type=str,
        default=None,
        help='Output directory of smoothed FITS image(s) [same as input file].')

    parser.add_argument(
        "--conv_mode",
        dest="conv_mode",
        type=str,
        default='robust',
        help="""Which method to use for convolution [robust].
        'robust' uses the built-in, FFT-based method.
        Can also be 'scipy', 'astropy', or 'astropy_fft'.
        Note these other methods cannot cope well with small convolving beams.
        """
    )

    parser.add_argument("-v", "--verbose", dest="verbose", action="store_true",
                        help="verbose output [False].")

    parser.add_argument("-d", "--dryrun", dest="dryrun", action="store_true",
                        help="Compute common beam and stop [False].")

    parser.add_argument(
        "--bmaj",
        dest="bmaj",
        type=float,
        default=None,
        help="Target BMAJ (arcsec) to convolve to [None].")

    parser.add_argument(
        "--bmin",
        dest="bmin",
        type=float,
        default=None,
        help="Target BMIN (arcsec) to convolve to [None].")

    parser.add_argument(
        "--bpa",
        dest="bpa",
        type=float,
        default=None,
        help="Target BPA (deg) to convolve to [None].")

    parser.add_argument(
        '-c',
        '--cutoff',
        dest='cutoff',
        type=float,
        default=None,
        help='Cutoff BMAJ value (arcsec) -- Blank channels with BMAJ larger than this [None -- no limit]')

    parser.add_argument(
        "-t",
        "--tolerance",
        dest="tolerance",
        type=float,
        default=0.0001,
        help="tolerance for radio_beam.commonbeam.")

    parser.add_argument(
        "-e",
        "--epsilon",
        dest="epsilon",
        type=float,
        default=0.0005,
        help="epsilon for radio_beam.commonbeam.")

    parser.add_argument(
        "-n",
        "--nsamps",
        dest="nsamps",
        type=int,
        default=200,
        help="nsamps for radio_beam.commonbeam.")

    group = parser.add_mutually_exclusive_group()

    group.add_argument("--ncores", dest="n_cores", default=1,
                       type=int, help="Number of processes (uses multiprocessing).")
    group.add_argument("--mpi", dest="mpi", default=False,
                       action="store_true", help="Run with MPI.")

    args = parser.parse_args()

    pool = schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores)
    if args.mpi:
        if not pool.is_master():
            pool.wait()
            sys.exit(0)

    verbose = args.verbose

    main(pool, args, verbose=verbose)
    pool.close()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
""" Convolve ASKAP cubes to common resolution """
__author__ = "Alec Thomson"

from .beamcon_2D import my_ceil, round_up
from spectral_cube.utils import SpectralCubeWarning
import warnings
from astropy.utils.exceptions import AstropyWarning
from astropy.convolution import convolve, convolve_fft
from . import convolve_uv
import os
import stat
import sys
import numpy as np
import scipy.signal
from astropy import units as u
from astropy.io import fits, ascii
from astropy.table import Table
from spectral_cube import SpectralCube
from radio_beam import Beam, Beams
from radio_beam.utils import BeamError
from tqdm import tqdm, trange
from . import au2
import functools
import psutil
from IPython import embed
try:
    from mpi4py import MPI
    mpiSwitch = True
except:
    mpiSwitch = False

# Fail if script has been started with mpiexec & mpi4py is not installed
if os.environ.get('OMPI_COMM_WORLD_SIZE') is not None:
    if not mpiSwitch:
        print("Script called with mpiexec, but mpi4py not installed")
        sys.exit()

# Get the processing environment
if mpiSwitch:
    comm = MPI.COMM_WORLD
    nPE = comm.Get_size()
    myPE = comm.Get_rank()
else:
    nPE = 1
    myPE = 0

print = functools.partial(print, f'[{myPE}]', flush=True)

warnings.filterwarnings(action='ignore', category=SpectralCubeWarning,
                        append=True)
warnings.simplefilter('ignore', category=AstropyWarning)

#############################################
#### ADAPTED FROM SCRIPT BY T. VERNSTROM ####
#############################################


class Error(OSError):
    pass


class SameFileError(Error):
    """Raised when source and destination are the same file."""


class SpecialFileError(OSError):
    """Raised when trying to do a kind of operation (e.g. copying) which is
    not supported on a special file (e.g. a named pipe)"""


class ExecError(OSError):
    """Raised when a command could not be executed"""


class ReadError(OSError):
    """Raised when an archive cannot be read"""


class RegistryError(Exception):
    """Raised when a registry operation with the archiving
    and unpacking registeries fails"""


def _samefile(src, dst):
    # Macintosh, Unix.
    if hasattr(os.path, 'samefile'):
        try:
            return os.path.samefile(src, dst)
        except OSError:
            return False


def copyfile(src, dst, *, follow_symlinks=True, verbose=True):
    """Copy data from src to dst.

    If follow_symlinks is not set and src is a symbolic link, a new
    symlink will be created instead of copying the file it points to.

    """
    if _samefile(src, dst):
        raise SameFileError("{!r} and {!r} are the same file".format(src, dst))

    for fn in [src, dst]:
        try:
            st = os.stat(fn)
        except OSError:
            # File most likely does not exist
            pass
        else:
            # XXX What about other special files? (sockets, devices...)
            if stat.S_ISFIFO(st.st_mode):
                raise SpecialFileError("`%s` is a named pipe" % fn)

    if not follow_symlinks and os.path.islink(src):
        os.symlink(os.readlink(src), dst)
    else:
        with open(src, 'rb') as fsrc:
            with open(dst, 'wb') as fdst:
                copyfileobj(fsrc, fdst, verbose=verbose)
    return dst


def copyfileobj(fsrc, fdst, length=16*1024, verbose=True):
    #copied = 0
    total = os.fstat(fsrc.fileno()).st_size
    with tqdm(
            total=total,
            disable=(not verbose),
            unit_scale=True,
            desc='Copying file'
    ) as pbar:
        while True:
            buf = fsrc.read(length)
            if not buf:
                break
            fdst.write(buf)
            copied = len(buf)
            pbar.update(copied)


def getbeams(beamlog, verbose=False):
    """

    colnames=['Channel', 'BMAJarcsec', 'BMINarcsec', 'BPAdeg']
    """
    # Get beamlog
    if verbose:
        print(f'Getting beams from {beamlog}')

    beams = Table.read(beamlog, format='ascii.commented_header')
    for col in beams.colnames:
        idx = col.find('[')
        if idx == -1:
            new_col = col
            unit = u.Unit('')
        else:
            new_col = col[:idx]
            unit = u.Unit(col[idx+1:-1])
        beams[col].unit = unit
        beams[col].name = new_col
    nchan = len(beams)
    return beams, nchan


def getfacs(datadict, convbeams, verbose=False):
    """Get smoothing unit factor

    Args:
        datadict (dict): Dict of input data
        convbeams (Beams): Convolving beams
        verbose (bool, optional): Verbose output. Defaults to False.

    Returns:
        list: factors to keep units of Jy/beam after convolution
    """    """Get beam info
    
    """
    facs = []
    for conbm, oldbeam in zip(convbeams, datadict['beams']):
        fac, amp, outbmaj, outbmin, outbpa = au2.gauss_factor(
            [
                conbm.major.to(u.arcsec).value,
                conbm.minor.to(u.arcsec).value,
                conbm.pa.to(u.deg).value
            ],
            beamOrig=[
                oldbeam.major.to(u.arcsec).value,
                oldbeam.minor.to(u.arcsec).value,
                oldbeam.pa.to(u.deg).value
            ],
            dx1=datadict['dx'].to(u.arcsec).value,
            dy1=datadict['dy'].to(u.arcsec).value
        )
        facs.append(fac)
    facs = np.array(facs)
    return facs


def smooth(image, dx, dy, oldbeam, newbeam, conbeam, sfactor,
           conv_mode='robust', verbose=False):
    """smooth an image in Jy/beam

    Args:
        image (ndarray): Image plane from FITS file
        dx (Quantity): deg per pixel in image
        dy (Quantity): deg per pixel in image
        oldbeam (Beam): Original PSF
        newbeam (Beam): Target PSF
        conbeam (Beam): Convolving beam
        sfactor (float): factor to keep units in Jy/beam
        conv_mode (str): Convolution mode
        verbose (bool, optional): Verbose output. Defaults to False.

    Returns:
        ndarray: Smoothed image
    """
    if np.isnan(conbeam):
        return image*np.nan
    if np.isnan(image).all():
        return image
    else:
        # using Beams package
        if verbose:
            print(f'Using convolving beam', conbeam)
            print(f'Using scaling factor', sfactor)
        pix_scale = dy
        gauss_kern = conbeam.as_kernel(pix_scale)

        conbm1 = gauss_kern.array/gauss_kern.array.max()
        fac = sfactor
        if conv_mode == 'robust':
            newim, fac = convolve_uv.convolve(
                image.astype('f8'),
                oldbeam,
                newbeam,
                dx,
                dy,
            )
        if conv_mode == 'scipy':
            newim = scipy.signal.convolve(
                image.astype('f8'),
                conbm1,
                mode='same'
            )
        elif conv_mode == 'astropy':
            newim = convolve(
                image.astype('f8'),
                conbm1,
                normalize_kernel=False,
            )
        elif conv_mode == 'astropy_fft':
            newim = convolve_fft(
                image.astype('f8'),
                conbm1,
                normalize_kernel=False,
                allow_huge=True,
            )
    if verbose:
        print(f'Using scaling factor', fac)
    newim *= fac
    return newim


def cpu_to_use(max_cpu, count):
    """Find number of cpus to use.
    Find the right number of cpus to use when dividing up a task, such
    that there are no remainders.
    Args:
        max_cpu (int): Maximum number of cores to use for a process.
        count (float): Number of tasks.

    Returns:
        Maximum number of cores to be used that divides into the number
        of tasks (int).
    """
    factors = []
    for i in range(1, count + 1):
        if count % i == 0:
            factors.append(i)
    factors = np.array(factors)
    return max(factors[factors <= max_cpu])


def worker(idx, cubedict, conv_mode='robust', start=0):
    """parallel worker function

    Args:
        idx (int): channel index
        cubedict (dict): Datadict referring to single image cube
        conv_mode (str): Convolution mode
        start (int, optional): index to start at. Defaults to 0.

    Returns:
        ndarray: smoothed image
    """
    cube = SpectralCube.read(cubedict["filename"])
    plane = cube.unmasked_data[start+idx].value
    newim = smooth(image=plane,
                   dx=cubedict['dx'],
                   dy=cubedict['dy'],
                   oldbeam=cubedict['beams'][start+idx],
                   newbeam=cubedict['commonbeams'][start+idx],
                   conbeam=cubedict['convbeams'][start+idx],
                   sfactor=cubedict['facs'][start+idx],
                   conv_mode=conv_mode,
                   verbose=False
                   )
    return newim


def makedata(files, outdir, verbose=True):
    """init datadict

    Args:
        files (list): list of input files
        outdir (list): list of output dirs

    Returns:
        datadict: Main data dictionary
    """
    datadict = {}
    for i, (file, out) in enumerate(zip(files, outdir)):
        # Set up files
        datadict[f"cube_{i}"] = {}
        datadict[f"cube_{i}"]["filename"] = file
        datadict[f"cube_{i}"]["outdir"] = out
        # Get metadata
        header = fits.getheader(file)
        dxas = header['CDELT1']*-1*u.deg
        datadict[f"cube_{i}"]["dx"] = dxas
        dyas = header['CDELT2']*u.deg
        datadict[f"cube_{i}"]["dy"] = dyas
        if not dxas == dyas:
            raise Exception("GRID MUST BE SAME IN X AND Y")
        # Get beam info
        dirname = os.path.dirname(file)
        basename = os.path.basename(file)
        if dirname == '':
            dirname = '.'
        beamlog = f"{dirname}/beamlog.{basename}".replace('.fits', '.txt')
        datadict[f"cube_{i}"]["beamlog"] = beamlog
        beam, nchan = getbeams(beamlog, verbose=verbose)
        datadict[f"cube_{i}"]["beam"] = beam
        datadict[f"cube_{i}"]["nchan"] = nchan
    return datadict


def commonbeamer(datadict, nchans, args, conv_mode='robust',
                 mode='natural', target_beam=None, verbose=True):
    """Find common beams

    Args:
        datadict (dict): Main data dict
        nchans (int): Number of channels
        args (args): Command line args
        conv_mode (str, optional): Convolution method
        mode (str, optional): 'total' or 'natural. Defaults to 'natural'.
        target_beam (Beam, optional): Target PSF
        verbose (bool, optional): Verbose output. Defaults to True.

    Returns:
        dict: updated datadict
    """
    ### Natural mode ###
    if mode == 'natural':
        big_beams = []
        for n in trange(
            nchans,
            desc='Constructing beams',
            disable=(not verbose)
        ):
            majors = []
            minors = []
            pas = []
            for key in datadict.keys():
                major = datadict[key]['beams'][n].major
                minor = datadict[key]['beams'][n].minor
                pa = datadict[key]['beams'][n].pa
                if datadict[key]['mask'][n]:
                    major *= np.nan
                    minor *= np.nan
                    pa *= np.nan
                majors.append(major.value)
                minors.append(minor.value)
                pas.append(pa.value)

            majors = np.array(majors)
            minors = np.array(minors)
            pas = np.array(pas)

            majors *= major.unit
            minors *= minor.unit
            pas *= pa.unit
            big_beams.append(Beams(major=majors, minor=minors, pa=pas))

        # Find common beams
        bmaj_common = []
        bmin_common = []
        bpa_common = []
        for beams in tqdm(
            big_beams,
            desc='Finding common beam per channel',
            disable=(not verbose),
            total=nchans
        ):
            if all(np.isnan(beams)):
                commonbeam = Beam(
                    major=np.nan*u.deg,
                    minor=np.nan*u.deg,
                    pa=np.nan*u.deg
                )
            else:
                try:
                    commonbeam = beams[~np.isnan(beams)].common_beam(tolerance=args.tolerance,
                                                                     nsamps=args.nsamps,
                                                                     epsilon=args.epsilon)
                except BeamError:
                    if verbose:
                        print("Couldn't find common beam with defaults")
                        print("Trying again with smaller tolerance")

                    commonbeam = beams[~np.isnan(beams)].common_beam(tolerance=args.tolerance*0.1,
                                                                     nsamps=args.nsamps,
                                                                     epsilon=args.epsilon)
                # Round up values
                commonbeam = Beam(
                    major=my_ceil(
                        commonbeam.major.to(u.arcsec).value, precision=1
                    )*u.arcsec,
                    minor=my_ceil(
                        commonbeam.minor.to(u.arcsec).value, precision=1
                    )*u.arcsec,
                    pa=round_up(commonbeam.pa.to(u.deg), decimals=2)
                )

                grid = datadict[key]["dy"]
                if conv_mode is not 'robust':
                    # Get the minor axis of the convolving beams
                    minorcons = []
                    for beam in beams[~np.isnan(beams)]:
                        minorcons += [commonbeam.deconvolve(
                            beam).minor.to(u.arcsec).value]
                    minorcons = np.array(minorcons)*u.arcsec
                    samps = minorcons / grid.to(u.arcsec)
                    # Check that convolving beam will be Nyquist sampled
                    if any(samps.value < 2):
                        # Set the convolving beam to be Nyquist sampled
                        nyq_con_beam = Beam(
                            major=grid*2,
                            minor=grid*2,
                            pa=0*u.deg
                        )
                        # Find new target based on common beam * Nyquist beam
                        # Not sure if this is best - but it works
                        nyq_beam = commonbeam.convolve(nyq_con_beam)
                        nyq_beam = Beam(
                            major=my_ceil(nyq_beam.major.to(
                                u.arcsec).value, precision=1)*u.arcsec,
                            minor=my_ceil(nyq_beam.minor.to(
                                u.arcsec).value, precision=1)*u.arcsec,
                            pa=round_up(nyq_beam.pa.to(u.deg), decimals=2)
                        )
                        if verbose:
                            print(
                                'Smallest common Nyquist sampled beam is:', nyq_beam)

                        warnings.warn('COMMON BEAM WILL BE UNDERSAMPLED!')
                        warnings.warn('SETTING COMMON BEAM TO NYQUIST BEAM')
                        commonbeam = nyq_beam

            bmaj_common.append(commonbeam.major.value)
            bmin_common.append(commonbeam.minor.value)
            bpa_common.append(commonbeam.pa.value)

        bmaj_common *= commonbeam.major.unit
        bmin_common *= commonbeam.minor.unit
        bpa_common *= commonbeam.pa.unit

        # Make Beams object
        commonbeams = Beams(
            major=bmaj_common,
            minor=bmin_common,
            pa=bpa_common
        )

    elif mode == 'total':
        majors = []
        minors = []
        pas = []
        for key in datadict.keys():
            major = datadict[key]['beams'].major
            minor = datadict[key]['beams'].minor
            pa = datadict[key]['beams'].pa
            major[datadict[key]['mask']] *= np.nan
            minor[datadict[key]['mask']] *= np.nan
            pa[datadict[key]['mask']] *= np.nan
            majors.append(major.value)
            minors.append(minor.value)
            pas.append(pa.value)

        majors = np.array(majors).ravel()
        minors = np.array(minors).ravel()
        pas = np.array(pas).ravel()

        majors *= major.unit
        minors *= minor.unit
        pas *= pa.unit
        big_beams = Beams(major=majors, minor=minors, pa=pas)

        if verbose:
            print('Finding common beam across all channels')
            print('This may take some time...')

        try:
            commonbeam = big_beams[~np.isnan(big_beams)].common_beam(tolerance=args.tolerance,
                                                                     nsamps=args.nsamps,
                                                                     epsilon=args.epsilon)
        except BeamError:
            if verbose:
                print("Couldn't find common beam with defaults")
                print("Trying again with smaller tolerance")

            commonbeam = big_beams[~np.isnan(big_beams)].common_beam(tolerance=args.tolerance*0.1,
                                                                     nsamps=args.nsamps,
                                                                     epsilon=args.epsilon)
        # Round up values
        commonbeam = Beam(
            major=my_ceil(
                commonbeam.major.to(u.arcsec).value, precision=1
            )*u.arcsec,
            minor=my_ceil(
                commonbeam.minor.to(u.arcsec).value, precision=1
            )*u.arcsec,
            pa=round_up(commonbeam.pa.to(u.deg), decimals=2)
        )
        if conv_mode is not 'robust':
            # Get the minor axis of the convolving beams
            grid = datadict[key]["dy"]
            minorcons = []
            for beam in big_beams[~np.isnan(big_beams)]:
                minorcons += [commonbeam.deconvolve(
                    beam).minor.to(u.arcsec).value]
            minorcons = np.array(minorcons)*u.arcsec
            samps = minorcons / grid.to(u.arcsec)
            # Check that convolving beam will be Nyquist sampled
            if any(samps.value < 2):
                # Set the convolving beam to be Nyquist sampled
                nyq_con_beam = Beam(
                    major=grid*2,
                    minor=grid*2,
                    pa=0*u.deg
                )
                # Find new target based on common beam * Nyquist beam
                # Not sure if this is best - but it works
                nyq_beam = commonbeam.convolve(nyq_con_beam)
                nyq_beam = Beam(
                    major=my_ceil(nyq_beam.major.to(
                        u.arcsec).value, precision=1)*u.arcsec,
                    minor=my_ceil(nyq_beam.minor.to(
                        u.arcsec).value, precision=1)*u.arcsec,
                    pa=round_up(nyq_beam.pa.to(u.deg), decimals=2)
                )
                if verbose:
                    print('Smallest common Nyquist sampled beam is:', nyq_beam)
                if target_beam is not None:
                    commonbeam = target_beam
                    if target_beam < nyq_beam:
                        warnings.warn('TARGET BEAM WILL BE UNDERSAMPLED!')
                        raise Exception("CAN'T UNDERSAMPLE BEAM - EXITING")
                else:
                    warnings.warn('COMMON BEAM WILL BE UNDERSAMPLED!')
                    warnings.warn('SETTING COMMON BEAM TO NYQUIST BEAM')
                    commonbeam = nyq_beam

        # Make Beams object
        commonbeams = Beams(
            major=[commonbeam.major] * nchans * commonbeam.major.unit,
            minor=[commonbeam.minor] * nchans * commonbeam.minor.unit,
            pa=[commonbeam.pa] * nchans * commonbeam.pa.unit
        )

    if verbose:
        print('Final beams are:')
        for i, commonbeam in enumerate(commonbeams):
            print(f'Channel {i}:', commonbeam)

    for key in tqdm(
        datadict.keys(),
        desc='Getting convolution data',
        disable=(not verbose)
    ):
        # Get convolving beams
        conv_bmaj = []
        conv_bmin = []
        conv_bpa = []
        oldbeams = datadict[key]['beams']
        masks = datadict[key]['mask']
        for commonbeam, oldbeam, mask in zip(commonbeams, oldbeams, masks):
            if mask:
                convbeam = Beam(
                    major=np.nan*u.deg,
                    minor=np.nan*u.deg,
                    pa=np.nan*u.deg
                )
            else:
                convbeam = commonbeam.deconvolve(oldbeam)
            conv_bmaj.append(convbeam.major.value)
            conv_bmin.append(convbeam.minor.value)
            conv_bpa.append(convbeam.pa.to(u.deg).value)

        conv_bmaj *= convbeam.major.unit
        conv_bmin *= convbeam.minor.unit
        conv_bpa *= u.deg

        # Construct beams object
        convbeams = Beams(
            major=conv_bmaj,
            minor=conv_bmin,
            pa=conv_bpa
        )

        # Get gaussian beam factors
        facs = getfacs(datadict[key], convbeams)
        datadict[key]['facs'] = facs

        # Setup conv beamlog
        datadict[key]['convbeams'] = convbeams
        commonbeam_log = datadict[key]['beamlog'].replace('beamlog.',
                                                          f'beamlogConvolve-{mode}.')
        datadict[key]['commonbeams'] = commonbeams
        datadict[key]['commonbeamlog'] = commonbeam_log

        commonbeam_tab = Table()
        # Save target
        commonbeam_tab.add_column(np.arange(nchans), name='Channel')
        commonbeam_tab.add_column(commonbeams.major, name='Target BMAJ')
        commonbeam_tab.add_column(commonbeams.minor, name='Target BMIN')
        commonbeam_tab.add_column(commonbeams.pa, name='Target BPA')
        # Save convolving beams
        commonbeam_tab.add_column(convbeams.major, name='Convolving BMAJ')
        commonbeam_tab.add_column(convbeams.minor, name='Convolving BMIN')
        commonbeam_tab.add_column(convbeams.pa, name='Convolving BPA')
        # Save facs
        commonbeam_tab.add_column(facs, name='Convolving factor')

        # Write to log file
        units = ''
        for col in commonbeam_tab.colnames:
            unit = commonbeam_tab[col].unit
            unit = str(unit)
            units += unit + ' '
        commonbeam_tab.meta['comments'] = [units]
        ascii.write(
            commonbeam_tab,
            output=commonbeam_log,
            format='commented_header',
            overwrite=True
        )
        if verbose:
            print(f'Convolving log written to {commonbeam_log}')

    return datadict


def masking(nchans, cutoff, datadict, verbose=True):
    for key in datadict.keys():
        mask = np.array([False]*nchans)
        datadict[key]['mask'] = mask
    if cutoff is not None:
        for key in datadict.keys():
            majors = datadict[key]['beams'].major
            cutmask = majors > cutoff
            datadict[key]['mask'] += cutmask

    # Check for pipeline masking
    nullbeam = Beam(major=0*u.deg, minor=0*u.deg, pa=0*u.deg)
    for key in datadict.keys():
        nullmask = datadict[key]['beams'] == nullbeam
        datadict[key]['mask'] += nullmask
    return datadict


def initfiles(datadict, mode, suffix=None, prefix=None, verbose=True):
    """Initialise output files

    Args:
        datadict (dict): Main data dict - indexed
        mode (str): 'total' or 'natural'
        verbose (bool, optional): Verbose output. Defaults to True.

    Returns:
        datadict: Updated datadict
    """
    with fits.open(datadict["filename"], memmap=True, mode='denywrite') as hdulist:
        primary_hdu = hdulist[0]
        data = primary_hdu.data
        header = primary_hdu.header

    # Header
    commonbeams = datadict['commonbeams']
    header = commonbeams[0].attach_to_header(header)
    primary_hdu = fits.PrimaryHDU(data=data, header=header)
    if mode == 'natural':
        header['COMMENT'] = 'The PSF in each image plane varies.'
        header['COMMENT'] = 'Full beam information is stored in the second FITS extension.'
        beam_table = Table(
            data=[
                commonbeams.major.to(u.arcsec),
                commonbeams.minor.to(u.arcsec),
                commonbeams.pa.to(u.deg)
            ],
            names=[
                'BMAJ',
                'BMIN',
                'BPA'
            ]
        )
        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        tab_hdu = fits.table_to_hdu(beam_table)
        new_hdulist = fits.HDUList([primary_hdu, tab_hdu])

    elif mode == 'total':
        new_hdulist = fits.HDUList([primary_hdu])

    # Set up output file
    if suffix is None:
        suffix = mode
    outname = os.path.basename(datadict["filename"])
    outname = outname.replace('.fits', f'.{suffix}.fits')
    if prefix is not None:
        outname = prefix + outname

    outdir = datadict['outdir']
    outfile = f'{outdir}/{outname}'
    if verbose:
        print(f'Initialising to {outfile}')

    new_hdulist.writeto(outfile, overwrite=True)

    return outfile


def readlogs(datadict, mode, verbose=True):
    if verbose:
        print('Reading from beamlogConvolve files')
    for key in datadict.keys():
        # Read in logs
        commonbeam_log = datadict[key]['beamlog'].replace('beamlog.',
                                                          f'beamlogConvolve-{mode}.')
        if verbose:
            print(f'Reading from {commonbeam_log}')
        try:
            commonbeam_tab = Table.read(
                commonbeam_log, format='ascii.commented_header')
        except FileNotFoundError:
            raise Exception("beamlogConvolve must be co-located with image")
        # Convert to Beams
        commonbeams = Beams(
            major=commonbeam_tab['Target BMAJ'] * u.arcsec,
            minor=commonbeam_tab['Target BMIN'] * u.arcsec,
            pa=commonbeam_tab['Target BPA'] * u.deg
        )
        convbeams = Beams(
            major=commonbeam_tab['Convolving BMAJ'] * u.arcsec,
            minor=commonbeam_tab['Convolving BMIN'] * u.arcsec,
            pa=commonbeam_tab['Convolving BPA'] * u.deg
        )
        facs = np.array(commonbeam_tab['Convolving factor'])
        # Save to datadict
        datadict[key]['facs'] = facs
        datadict[key]['convbeams'] = convbeams
        datadict[key]['commonbeams'] = commonbeams
        datadict[key]['commonbeamlog'] = commonbeam_log
    if verbose:
        print('Final beams are:')
        for i, commonbeam in enumerate(commonbeams):
            print(f'Channel {i}:', commonbeam)
    return datadict


def main(args, verbose=True):
    """main script

    Args:
        args (args): Command line args
        verbose (bool, optional): Verbose ouput. Defaults to True.

    """

    if myPE == 0:
        print(f"Total number of MPI ranks = {nPE}")
        # Parse args
        if args.dryrun:
            if verbose:
                print('Doing a dry run -- no files will be saved')

        # Check mode
        mode = args.mode
        if verbose:
            print(f"Mode is {mode}")
        if mode == 'natural' and mode == 'total':
            raise Exception("'mode' must be 'natural' or 'total'")
        if mode == 'natural':
            if verbose:
                print('Smoothing each channel to a common resolution')
        if mode == 'total':
            if verbose:
                print('Smoothing all channels to a common resolution')

        # Check cutoff
        cutoff = args.cutoff
        if args.cutoff is not None:
            cutoff = args.cutoff * u.arcsec
            if verbose:
                print('Cutoff is:', cutoff)

        # Check target
        conv_mode = args.conv_mode
        print(conv_mode)
        if not conv_mode == 'robust' and not conv_mode == 'scipy' and \
                not conv_mode == 'astropy' and not conv_mode == 'astropy_fft':
            raise Exception('Please select valid convolution method!')

        if verbose:
            print(f"Using convolution method {conv_mode}")
            if conv_mode == 'robust':
                print("This is the most robust method. And fast!")
            elif conv_mode == 'scipy':
                print('This fast, but not robust to NaNs or small PSF changes')
            else:
                print('This is slower, but robust to NaNs, but not to small PSF changes')

        bmaj = args.bmaj
        bmin = args.bmin
        bpa = args.bpa

        nonetest = [test is None for test in [bmaj, bmin, bpa]]

        if not all(nonetest) and mode is not 'total':
            raise Exception("Only specify a target beam in 'total' mode")

        if all(nonetest):
            target_beam = None

        elif not all(nonetest) and any(nonetest):
            raise Exception('Please specify all target beam params!')

        elif not all(nonetest) and not any(nonetest):
            target_beam = Beam(
                bmaj * u.arcsec,
                bmin * u.arcsec,
                bpa * u.deg
            )
            if verbose:
                print('Target beam is ', target_beam)

        files = sorted(args.infile)
        if files == []:
            raise Exception('No files found!')

        outdir = args.outdir
        if outdir is not None:
            if outdir[-1] == '/':
                outdir = outdir[:-1]
            outdir = [outdir] * len(files)
        else:
            outdir = []
            for f in files:
                out = os.path.dirname(f)
                if out == '':
                    out = '.'
                outdir += [out]

        datadict = makedata(files, outdir, verbose=verbose)

        # Sanity check channel counts
        nchans = np.array([datadict[key]['nchan'] for key in datadict.keys()])
        check = all(nchans == nchans[0])

        if not check:
            raise Exception('Unequal number of spectral channels!')

        else:
            nchans = nchans[0]

        # Construct Beams objects
        for key in datadict.keys():
            beam = datadict[key]['beam']
            bmaj = np.array(beam['BMAJ'])*beam['BMAJ'].unit
            bmin = np.array(beam['BMIN'])*beam['BMIN'].unit
            bpa = np.array(beam['BPA'])*beam['BPA'].unit
            beams = Beams(
                major=bmaj,
                minor=bmin,
                pa=bpa
            )
            datadict[key]['beams'] = beams

        # Apply some masking
        datadict = masking(
            nchans,
            cutoff,
            datadict,
            verbose=verbose
        )

        if not args.uselogs:
            datadict = commonbeamer(
                datadict,
                nchans,
                args,
                conv_mode=conv_mode,
                target_beam=target_beam,
                mode=mode,
                verbose=verbose
            )
        else:
            datadict = readlogs(
                datadict,
                mode=mode,
                verbose=verbose
            )

    else:
        if not args.dryrun:
            files = None
            datadict = None
            nchans = None

    if mpiSwitch:
        comm.Barrier()

    # Init the files in parallel
    if not args.dryrun:
        if myPE == 0 and verbose:
            print('Initialising output files')
        if mpiSwitch:
            files = comm.bcast(files, root=0)
            datadict = comm.bcast(datadict, root=0)
            nchans = comm.bcast(nchans, root=0)

        conv_mode = args.conv_mode
        inputs = list(datadict.keys())
        dims = len(inputs)

        if nPE > dims:
            my_start = myPE
            my_end = myPE

        else:
            count = dims // nPE
            rem = dims % nPE
            if myPE < rem:
                # The first 'remainder' ranks get 'count + 1' tasks each
                my_start = myPE * (count + 1)
                my_end = my_start + count

            else:
                # The remaining 'size - remainder' ranks get 'count' task each
                my_start = myPE * count + rem
                my_end = my_start + (count - 1)

        if verbose:
            if myPE == 0:
                print(
                    f"There are {dims} files to init")
            print(f"My start is {my_start}", f"My end is {my_end}")

        # Init output files and retrieve file names
        outfile_dict = {}
        for inp in inputs[my_start:my_end+1]:
            outfile = initfiles(
                datadict[inp],
                args.mode,
                suffix=args.suffix,
                prefix=args.prefix,
                verbose=verbose
            )
            outfile_dict.update(
                {
                    inp: outfile
                }
            )

        if mpiSwitch:
            # Send to master proc
            outlist = comm.gather(outfile_dict, root=0)

        if mpiSwitch:
            comm.Barrier()

        # Now do the convolution in parallel
        if myPE == 0:

            # Conver list to dict and save to main dict
            outlist_dict = {}
            for d in outlist:
                outlist_dict.update(d)
            # Also make inputs list
            inputs = []
            for key in datadict.keys():
                datadict[key]['outfile'] = outlist_dict[key]
                for chan in range(nchans):
                    inputs.append((key, chan))

        else:
            datadict = None
            inputs = None
        if mpiSwitch:
            comm.Barrier()
        if mpiSwitch:
            inputs = comm.bcast(inputs, root=0)
            datadict = comm.bcast(datadict, root=0)

        dims = len(files) * nchans
        assert len(inputs) == dims
        count = dims // nPE
        rem = dims % nPE
        if myPE < rem:
            # The first 'remainder' ranks get 'count + 1' tasks each
            my_start = myPE * (count + 1)
            my_end = my_start + count

        else:
            # The remaining 'size - remainder' ranks get 'count' task each
            my_start = myPE * count + rem
            my_end = my_start + (count - 1)
        if verbose:
            if myPE == 0:
                print(
                    f"There are {nchans} channels, across {len(files)} files")
            print(f"My start is {my_start}", f"My end is {my_end}")

        for inp in inputs[my_start:my_end+1]:
            key, chan = inp
            newim = worker(chan, datadict[key], conv_mode=conv_mode)
            outfile = datadict[key]['outfile']
            with fits.open(outfile, mode='update', memmap=True) as outfh:
                outfh[0].data[chan, 0, :, :] = newim
                outfh.flush()
            if verbose:
                print(f"{outfile}  - channel {chan} - Done")

    if verbose:
        print('Done!')


def cli():
    """Command-line interface
    """
    import argparse

    # Help string to be shown using the -h option
    descStr = """
    Smooth a field of 3D cubes to a common resolution.

    Names of output files are 'infile'.sm.fits

    """

    # Parse the command line options
    parser = argparse.ArgumentParser(description=descStr,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        'infile',
        metavar='infile',
        type=str,
        help="""Input FITS image(s) to smooth (can be a wildcard) 
        - beam info must be in co-located beamlog files.
        """,
        nargs='+')

    parser.add_argument(
        "--uselogs",
        dest="uselogs",
        action="store_true",
        help="Get convolving information from previous run [False]."
    )

    parser.add_argument(
        '--mode',
        dest='mode',
        type=str,
        default='natural',
        help="""Common resolution mode [natural]. 
        natural  -- allow frequency variation.
        total -- smooth all plans to a common resolution.
        """
    )

    parser.add_argument(
        "--conv_mode",
        dest="conv_mode",
        type=str,
        default='robust',
        help="""Which method to use for convolution [robust].
        'robust' uses the built-in, FFT-based method.
        Can also be 'scipy', 'astropy', or 'astropy_fft'.
        Note these other methods cannot cope well with small convolving beams.
        """
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="verbose output [False]."
    )

    parser.add_argument(
        "-d",
        "--dryrun",
        dest="dryrun",
        action="store_true",
        help="Compute common beam and stop [False]."
    )

    parser.add_argument(
        '-p',
        '--prefix',
        dest='prefix',
        type=str,
        default=None,
        help='Add prefix to output filenames.')

    parser.add_argument(
        '-s',
        '--suffix',
        dest='suffix',
        type=str,
        default=None,
        help='Add suffix to output filenames [...{mode}.fits].')

    parser.add_argument(
        '-o',
        '--outdir',
        dest='outdir',
        type=str,
        default=None,
        help='Output directory of smoothed FITS image(s) [None - same as input].'
    )

    parser.add_argument(
        "--bmaj",
        dest="bmaj",
        type=float,
        default=None,
        help="BMAJ to convolve to [max BMAJ from given image(s)]."
    )

    parser.add_argument(
        "--bmin",
        dest="bmin",
        type=float,
        default=None,
        help="BMIN to convolve to [max BMAJ from given image(s)]."
    )

    parser.add_argument(
        "--bpa",
        dest="bpa",
        type=float,
        default=None,
        help="BPA to convolve to [0]."
    )

    parser.add_argument(
        '-m',
        '--mask',
        dest='masklist',
        type=str,
        default=None,
        help='List of channels to be masked [None]'
    )

    parser.add_argument(
        '-c',
        '--cutoff',
        dest='cutoff',
        type=float,
        default=None,
        help='Cutoff BMAJ value (arcsec) -- Blank channels with BMAJ larger than this [None -- no limit]'
    )

    parser.add_argument(
        "-t",
        "--tolerance",
        dest="tolerance",
        type=float,
        default=0.0001,
        help="tolerance for radio_beam.commonbeam."
    )

    parser.add_argument(
        "-e",
        "--epsilon",
        dest="epsilon",
        type=float,
        default=0.0005,
        help="epsilon for radio_beam.commonbeam."
    )

    parser.add_argument(
        "-n",
        "--nsamps",
        dest="nsamps",
        type=int,
        default=200,
        help="nsamps for radio_beam.commonbeam."
    )

    args = parser.parse_args()

    verbose = args.verbose

    main(args, verbose=verbose)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python

import functools
import astropy.units as u
import numpy as np
from spectral_cube import SpectralCube
from tqdm import tqdm, trange
import schwimmbad
import time
import warnings
import sys
from spectral_cube.utils import SpectralCubeWarning
warnings.filterwarnings(action='ignore', category=SpectralCubeWarning,
                        append=True)
import psutil
print = functools.partial(print,f'[{psutil.Process().cpu_num()}]', flush=True)


#############################################
####### ADAPTED FROM SCRIPT BY G. HEALD #####
#############################################

def myfit(x, y, fn):
    # Find the width of a Gaussian distribution by computing the second moment of
    # the data (y) on a given axis (x)

    w = np.sqrt(abs(sum(x**2*y)/sum(y)))

    # Or something like this. Probably a Gauss plus a power-law with a lower cutoff is necessary
    #func = lambda x, a, b: np.exp(0.5*x**2/a**2) + x**(-1.*b)

    #[popt, pvar] = curve_fit(func, x, y)
    # Need a newer version of scipy for this to work...

    #a = popt[0]
    #b = popt[1]
    if fn != '':
        import matplotlib.pyplot as plt
        plt.ioff()
        plt.semilogy(x, y, '+', label="Data")
        #plt.semilogy(x, np.exp(-0.5*x**2/a**2) + x**(-1.*b), label="Noise fit")
        plt.semilogy(x, np.exp(-0.5*x**2/w**2), label="Noise fit")
        plt.title('Normalized pixel distribution')
        plt.ylabel('Rel. Num. of Pixels')
        plt.xlabel('Pixel brightness (Jy/beam)')
        plt.legend(loc=3)
        plt.savefig(fn)
        # plt.show()
        plt.close()
    return w


def calcnoise(args):
    """Get noise in plane from cube.
    """
    i, file, totalbad, update = args
    if update:
        print(f'Checking channel {i}')
    if totalbad is not None and totalbad[i]:
        return -1
    else:
        cube = getcube(file)
        plane = cube.unmasked_data[i]
        imsize = plane.shape
        assert len(imsize) == 2
        nx = imsize[-1]
        ny = imsize[-2]
        Id = plane[ny//3:2*ny//3, nx//3:2*nx//3].flatten()
        if len(Id[np.isnan(Id)]) == len(Id):
            return -1.
        else:
            rms = np.std(Id)
            mval = np.mean(Id)
            Id = Id[np.logical_and(Id < mval+3.*rms, Id > mval-3.*rms)]
            # print mval,rms,len(Id)

            #hrange = (-1,1)
            # , range=hrange) # 0 = values, 1 = left bin edges
            Ih = np.histogram(Id, bins=100)
            if max(Ih[0]) == 0.:
                return -1.
            Ix = Ih[1][:-1] + 0.5*(Ih[1][1] - Ih[1][0])
            Iv = Ih[0]/float(max(Ih[0]))
            Inoise = myfit(Ix, Iv, '')
            return Inoise

def getcube(filename):
    """Read FITS file as SpectralCube

    Masks out 0Jy/beam pixels

    """
    cube = SpectralCube.read(filename)
    mask = ~(cube == 0*u.jansky/u.beam)
    cube = cube.with_mask(mask)
    return cube


def getbadchans(pool, qcube, ucube, ufile, qfile, totalbad=None, cliplev=5, update=False):
    """Find deviated channels
    """
    assert len(ucube.spectral_axis) == len(qcube.spectral_axis)
    inputs = [[i, qfile, totalbad, update]
              for i in range(len(qcube.spectral_axis))]
    if (pool.__class__.__name__ is 'MPIPool' or
            pool.__class__.__name__ is 'SerialPool'):
        print(f'Checking Q...')
        tic = time.perf_counter()
        qnoisevals = list(
            pool.map(calcnoise, inputs)
        )
        toc = time.perf_counter()
        print(f'Time taken was {toc - tic}s')

    elif pool.__class__.__name__ is 'MultiPool':
        qnoisevals = list(tqdm(
            pool.imap_unordered(calcnoise, inputs),
            total=len(ucube.spectral_axis),
            desc='Checking Q',
        )
        )
    qnoisevals = np.array(qnoisevals)

    inputs = [[i, ufile, totalbad, update]
              for i in range(len(ucube.spectral_axis))]
    if (pool.__class__.__name__ is 'MPIPool' or
            pool.__class__.__name__ is 'SerialPool'):
        print(f'Checking U...')
        tic = time.perf_counter()
        unoisevals = list(
            pool.map(calcnoise, inputs)
        )
        toc = time.perf_counter()
        print(f'Time taken was {toc - tic}s')

    elif pool.__class__.__name__ is 'MultiPool':
        unoisevals = list(tqdm(
            pool.imap_unordered(calcnoise, inputs),
            total=len(ucube.spectral_axis),
            desc='Checking U',
        )
        )
    unoisevals = np.array(unoisevals)
    qmeannoise = np.median(qnoisevals[abs(qnoisevals) < 1.])
    qstdnoise = np.std(qnoisevals[abs(qnoisevals) < 1.])
    print('Q median, std:', qmeannoise, qstdnoise)
    umeannoise = np.median(unoisevals[abs(unoisevals) < 1.])
    ustdnoise = np.std(unoisevals[abs(unoisevals) < 1.])
    print('U median, std:', umeannoise, ustdnoise)
    qbadones = np.logical_or(qnoisevals > (
        qmeannoise+cliplev*qstdnoise), qnoisevals == -1.)
    ubadones = np.logical_or(unoisevals > (
        umeannoise+cliplev*ustdnoise), unoisevals == -1.)
    print(sum(np.asarray(qbadones, dtype=int)),
          'of', len(qcube.spectral_axis), 'are bad (Q)')
    print(sum(np.asarray(ubadones, dtype=int)),
          'of', len(ucube.spectral_axis), 'are bad (U)')
    totalbad = np.logical_or(qbadones, ubadones)
    print(sum(np.asarray(totalbad, dtype=int)), 'of',
          len(qcube.spectral_axis), 'are bad in Q -or- U')
    return totalbad


def blankchans(qcube, ucube, totalbad, blank=False):
    """Mask out bad chans
    """
    chans = np.array([i for i, chan in enumerate(qcube.spectral_axis)])
    badchans = chans[totalbad]
    badfreqs = qcube.spectral_axis[totalbad]
    if not blank:
        print('Nothing will be blanked, but these are the channels/frequencies that would be with the -b option activated:')
    print(f'Bad channels are {badchans}')
    print(f'Bad frequencies are {badfreqs}')
    totalgood = [not bad for bad in totalbad]
    q_msk = qcube.mask_channels(totalgood)
    u_msk = ucube.mask_channels(totalgood)
    if not args.blank:
        print('Nothing in the above list was blanked, use -b to take that action')
    return q_msk, u_msk


def writefits(qcube, ucube, clargs):
    """Write output to disk
    """
    outfile = clargs.qfitslist.replace('.fits', '.blanked.fits')
    print(f'Writing to {outfile}')
    qcube.write(outfile, format='fits', overwrite=True)
    outfile = clargs.ufitslist.replace('.fits', '.blanked.fits')
    print(f'Writing to {outfile}')
    ucube.write(outfile, format='fits', overwrite=True)


def main(pool, clargs):
    qcube = getcube(clargs.qfitslist)
    ucube = getcube(clargs.ufitslist)

    totalbad = None

    if clargs.mpi or clargs.n_cores == 1:
        update = True
    else:
        update = False

    totalbad = getbadchans(pool, qcube, ucube, clargs.qfitslist,
                           clargs.ufitslist, totalbad=totalbad, cliplev=clargs.cliplev, update=update)
    # print fitslist[np.asarray(badones,dtype=int)]

    q_msk, u_msk = blankchans(qcube, ucube, totalbad, blank=clargs.blank)

    if clargs.iterate is not None:
        print(f'Iterating {clargs.iterate} additional time(s)...')
        for i in range(clargs.iterate):
            totalbad = getbadchans(pool, q_msk, u_msk, clargs.qfitslist,
                                   clargs.ufitslist, totalbad=totalbad, cliplev=clargs.cliplev, update=update)
            q_msk, u_msk = blankchans(
                q_msk, u_msk, totalbad, blank=clargs.blank)

    if clargs.blank:
        writefits(q_msk, u_msk, clargs)

    if clargs.file is not None:
        print(f'Saving bad files to {clargs.file}')
        np.savetxt(clargs.file, totalbad)

def cli():
    import argparse
    descStr = """
    Find bad channels by checking statistics of each channel image.

    """
    warnings.filterwarnings(
        "ignore", message="Cube is a Stokes cube, returning spectral cube for I component")
    warnings.filterwarnings(
        "ignore", message="Invalid value encountered in less")
    warnings.filterwarnings(
        "ignore", message="Invalid value encountered in greater")
    warnings.filterwarnings(
        "ignore", message="overflow encountered in square")

    # Parse the command line options
    ap = argparse.ArgumentParser(description=descStr,
                                 formatter_class=argparse.RawTextHelpFormatter)
    ap.add_argument('qfitslist', help='Wildcard list of Q fits files')
    ap.add_argument('ufitslist', help='Wildcard list of U fits files')
    ap.add_argument(
        '--blank',
        '-b',
        help='Blank bad channel maps? [default False]',
        default=False,
        action='store_true'
    )
    ap.add_argument(
        '--cliplev',
        '-c',
        help='Clip level in sigma, make this number lower to be more aggressive [default 5]',
        default=5.,
        type=float
    )

    ap.add_argument(
        '--iterate',
        '-i',
        help='Iterate flagging check N additional times [None -- one pass only]',
        default=None,
        type=int
    )

    ap.add_argument(
        '--file',
        '-f',
        help='Filename to write bad channel indices to file [None --  do not write]',
        default=None,
        type=str
    )

    group = ap.add_mutually_exclusive_group()

    group.add_argument("--ncores", dest="n_cores", default=1,
                       type=int, help="Number of processes (uses multiprocessing).")
    group.add_argument("--mpi", dest="mpi", default=False,
                       action="store_true", help="Run with MPI.")

    args = ap.parse_args()

    pool = schwimmbad.choose_pool(
        mpi=args.mpi, processes=args.n_cores)
    if args.mpi:
        if not pool.is_master():
            pool.wait()
            sys.exit(0)

    main(pool, args)

    pool.close()

if __name__ == "__main__":
    cli()



#!/usr/bin/env python
""" For getting fluxes right in Jy/beam """
__author__ = "Tessa Vernstrom"

from scipy import *
import numpy as np
import math



def gaussianDeconvolve(smaj, smin, spa, bmaj, bmin, bpa):
    """'s' as in 'source', 'b' as in 'beam'. All arguments in
    radians. (Well, major and minor axes can be in any units, so long
    as they're consistent.)

    Returns dmaj, dmin, dpa, status
    Return units are consistent with the inputs.
    status is one of 'ok', 'pointlike', 'fail'

    Derived from miriad gaupar.for:GauDfac()

    We currently don't do a great job of dealing with pointlike
    sources. I've added extra code ensure smaj >= bmaj, smin >= bmin,
    and increased coefficient in front of "limit" from 0.1 to
    0.5. Feel a little wary about that first change.
    """

    from numpy import cos, sin, sqrt, min, abs, arctan2
    import numpy as np
    
    spa=np.radians(spa)
    bpa=np.radians(bpa)
    if smaj < bmaj:
        smaj = bmaj
    if smin < bmin:
        smin = bmin

    alpha = ((smaj * cos (spa))**2 + (smin * sin (spa))**2 -
             (bmaj * cos (bpa))**2 - (bmin * sin (bpa))**2)
    beta = ((smaj * sin (spa))**2 + (smin * cos (spa))**2 -
            (bmaj * sin (bpa))**2 - (bmin * cos (bpa))**2)
    gamma = 2 * ((smin**2 - smaj**2) * sin (spa) * cos (spa) -
                 (bmin**2 - bmaj**2) * sin (bpa) * cos (bpa))
#    print smaj,smin
#    print alpha,beta,gamma
    s = alpha + beta
    t = sqrt ((alpha - beta)**2 + gamma**2)
#    print s,t
    dmaj = sqrt (0.5 * (s + t))
    if s>t:
        dmin = sqrt (0.5 * (s - t))
    else:
        dmin= 0
#    print dmaj,dmin
    if alpha < 0 or beta < 0:
        dmaj = dmin = dpa = 0
    
#    if(smaj>bmaj):
#        dmaj= sqrt (0.5 * (s + t))
    if abs (gamma) + abs (alpha - beta) == 0:
        dpa = 0
    else:
        dpa=0.5 * arctan2 (-gamma, alpha - beta)
#    if((s>=t)&(bmin!=smin)):
#        dmin=sqrt (0.5 * (s - t))

    return dmaj, dmin, np.degrees(dpa)





def gauss_factor(beamConv, beamOrig=None, dx1=1, dy1=1):
    """
    Calculates the scaling factor to be applied after convolving
    a map in Jy/beam with a gaussian to get fluxes in Jy/beam again.

    This function is a copy of the FORTRAN gaufac function from the Miriad
    package, which determine the Gaussian parameters resulting from
    convolving two gaussians. This function yields the same result as
    the MIRIAD gaufac function.

    Parameters
    ----------
    beamConv : list
        A list of the [major axis, minor axis, position_angle]
        of the gaussion used for convolution.
    beamOrig :
        Same format as beamConv but giving the parameters of the original
        beam of the map. As Default the self.resolution list is used.
    dx1, dy1 : floats
        Being the pixel size in both dimensions of the map.
        By default the ``CDELT1`` and ``CDELT2`` keywords from the
        fits header are used.

    Returns
    -------
    fac :
        Factor for the output Units.
    amp :
        Amplitude of resultant gaussian.
    bmaj, bmin :
        Major and minor axes of resultant gaussian.
    bpa :
        Position angle of the resulting gaussian.
    """
    # include 'mirconst.h'
    # Define cosine and Sinus of the position Angles of the
    # Gaussians
    arcsecInGrad=1#(1./3600)*(np.pi/180.)
    deg2Grad=(np.pi/180)
    bmaj2, bmin2, bpa2 = beamConv
    bmaj2, bmin2, bpa2 = (bmaj2 * arcsecInGrad, bmin2 *
                          arcsecInGrad, bpa2 * deg2Grad)
    #if beamOrig is None:
    bmaj1, bmin1, bpa1 = beamOrig
    bmaj1, bmin1, bpa1 = (bmaj1 * arcsecInGrad,
                              bmin1 * arcsecInGrad,
                              bpa1 * deg2Grad)
    #if dx1 is None:
    dx1 = dx1 * arcsecInGrad
    #if dy1 is None:
    dy1 = dy1 * arcsecInGrad
    cospa1 = math.cos(bpa1)
    cospa2 = math.cos(bpa2)
    sinpa1 = math.sin(bpa1)
    sinpa2 = math.sin(bpa2)
    alpha = ((bmaj1 * cospa1) ** 2
             + (bmin1 * sinpa1) ** 2
             + (bmaj2 * cospa2) ** 2
             + (bmin2 * sinpa2) ** 2)
    beta = ((bmaj1 * sinpa1) ** 2
            + (bmin1 * cospa1) ** 2
            + (bmaj2 * sinpa2) ** 2
            + (bmin2 * cospa2) ** 2)
    gamma = (2 * ((bmin1 ** 2 - bmaj1 ** 2)
                  * sinpa1 * cospa1
                  + (bmin2 ** 2 - bmaj2 ** 2)
                  * sinpa2 * cospa2))
    s = alpha + beta
    t = math.sqrt((alpha - beta) ** 2 + gamma ** 2)
    bmaj = math.sqrt(0.5 * (s + t))
    bmin = math.sqrt(0.5 * (s - t))
    if (abs(gamma) + abs(alpha - beta)) == 0:
        bpa = 0.0
    else:
        bpa = 0.5 * np.arctan2(-1 * gamma, alpha - beta)
        #print alpha,beta,gamma
    amp = (math.pi / (4.0 * math.log(2.0)) * bmaj1 * bmin1 * bmaj2 * bmin2
           / math.sqrt(alpha * beta - 0.25 * gamma * gamma))
    fac = ((math.sqrt(dx1 ** 2) * math.sqrt(dy1 ** 2))) / amp

    return fac, amp, bmaj , bmin , np.degrees(bpa)



#!/usr/bin/env python
""" Fast convolution in the UV domain """
__author__ = "Wasim Raja"

import numpy as np
import astropy.units as units
import racs_tools.gaussft as gaussft


def convolve(image, old_beam, new_beam, dx, dy):
    """Convolve in the UV domain.

    Args:
        image (2D array): The image to be convolved.
        old_beam (radio_beam.Beam): Current image PSF.
        new_beam (radio_beam.Beam): Target image PSF.
        dx (float): Grid size in x in degrees (e.g. CDELT1)
        dy (float): Grid size in y in degrees (e.g. CDELT2)

    Returns:
        tuple: (conolved image, scaling factor)
    """
    nx = image.shape[0]
    ny = image.shape[1]

    # The coordinates in FT domain:
    u = np.fft.fftfreq(nx, d=dx.to(units.rad).value)
    v = np.fft.fftfreq(ny, d=dy.to(units.rad).value)

    g_final = np.zeros((nx, ny), dtype=float)
    [g_final, g_ratio] = gaussft.gaussft(bmin_in=old_beam.minor.to(units.deg).value,
                                         bmaj_in=old_beam.major.to(
                                             units.deg).value,
                                         bpa_in=old_beam.pa.to(
                                             units.deg).value,
                                         bmin=new_beam.minor.to(
                                             units.deg).value,
                                         bmaj=new_beam.major.to(
                                             units.deg).value,
                                         bpa=new_beam.pa.to(units.deg).value,
                                         u=u, v=v,
                                         nx=nx, ny=ny)
    # Perform the x-ing in the FT domain
    im_f = np.fft.fft2(image)

    # Now convolve with the desired Gaussian:
    M = np.multiply(im_f, g_final)
    im_conv = np.fft.ifft2(M)
    im_conv = np.real(im_conv)

    # print("factor: %f" % g_ratio)
    # print("dx: %s" % dx)
    # print("dy: %s" % dy)
    # tmp = old_beam.minor.to(units.deg).value
    # print("bMaj psf: %f , %f" % (tmp,tmp*3600))
    # tmp = bmin_in
    # print("bMaj psf: %f , %f" % (tmp,tmp*3600.0))
    # tmp = bpa_in
    # print("bPA psf: %f " % tmp)
    # tmp = bmaj
    # print("bMaj desired: %f, %f" % (tmp,tmp*3600.0))
    # tmp = bmin
    # print("bMin desired: %f, %f" % (tmp,tmp*3600.0))
    # tmp = bpa
    # print("bPA desired: %f " % tmp)
    return im_conv, g_ratio
